# frozen_string_literal: trues

require 'timeout'

module Hermes
  # HousingIntention model
  class HousingApplication
    include Hermes::Concerns::CommonFieldsConcern
    include Hermes::Concerns::Status
    include Mongoid::Attributes::Dynamic
    
    include Hermes::Concerns::SopraXmlSender
    field :applicated_at, type: DateTime
    field :effort_rate, type: Float
    field :epci, type: Integer
    field :is_active, type: Boolean, default: true
    field :max_date, type: DateTime
    field :visit_conducted_date, type: DateTime
    field :date_event, type: Date
    field :last_status_update_date, type: Date
    field :lessor_event_informations, type: String
    field :sopra_code, type: String
    field :als_id, type: String
    field :rank, type: Integer
    
    field :profile_score, type: Integer
    field :eligible_position, type: Integer
    field :applicated_position, type: Integer
    field :effort_rate, type: Float

    field :cotations, type: Hash
    field :business_cotation, type: Float
    
    common_indexes *%i[
      applicated_at
      effort_rate
      epci
      is_active
      max_date
      visit_conducted_date
      last_status_update_date
      lessor_event_informations
      sopra_code
      rank
      profile_score
      eligible_position
      applicated_position
      effort_rate
      als_id
      cotations
    ]
    
    belongs_to :housing_offer, index: true, class_name: 'Hermes::HousingOffer'
    belongs_to :housing_request, index: true, class_name: 'Hermes::HousingRequest'
    
    validates_presence_of :housing_request_id, :housing_offer, unless: -> {disabled_or_deleted_status?}
    validates :status, inclusion: {in: Hermes::Statuses::HousingApplication.possible_status.values.flatten, allow_blank: true}
    validates :visit_status, inclusion: {in: Hermes::Statuses::HousingApplication.possible_status[:visit_status], allow_blank: true}
    validate :validate_request_supplement
    
    writer_auth_roles "collaborateur offre logement", 'superadministrateur'
    writer_roles 'Superadmin', 'Administrator'
    reader_roles 'Manager'
    owner_field :user_id
    
    before_validation :before_validation_actions
    before_create :before_create_actions
    before_save :before_save_actions
    after_save :after_save_actions
    after_create :after_create_actions
    after_update :after_actions
    after_destroy :after_destroy_actions
    
    scope :selected_by_offer, ->(housing_offer_id) {
      where(housing_offer_id: housing_offer_id, status: "TRANSMIS BAILLEUR")
    }
    
    scope :actives, ->() {where(is_active: true)}
    
    HousingApplication.add_status(:single)
    HousingApplication.add_status(:single, "visit_status")
    
    ###############################
    #### Notification #############
    include ::Concerns::NotificationCenter::Notifiable

    def self.targets_selector(housing_app, opt = {})
      [housing_app.applicant].compact
    end
    
    def notification_mailer
      ::Hermes::NotificationCenter::HermesMailer
    end
    
    def notifiable_title(key, opt = {})
      I18n.t("#{key}.title")
    end
    
    def notifiable_body(key, opt = {})
      I18n.t("#{key}.body", number: housing_offer.reference, fullname: applicant.full_name)
    end
    
    def notifiable_path
      "https://als-hermes-int.be-idys.com/housing_applications/#{id}"
    end
    
    ###############################
    ###############################

    def validate_request_supplement
      hr = Hermes::HousingRequest.find(housing_request_id)
      return true unless self.new_record? && hr&.is_provider_imhoweb
      
      begin
        Timeout::timeout(10) do
          sleep(HERMES_CONFIG['sleeptimeout']) if HERMES_CONFIG['sleeptimeout'] # for testing
          hr.update_request_supplement(user)

          if hr.supplement_can_apply == false
            # TODO : remove when imhoweb fixed
            # self.errors.add(:housing_request_id, "this housing_request is not allowed to apply")
            # return false
            return true
          end
          
          return true
        end
      rescue Timeout::Error
        return true
      end

    end

    
    def to_json(options = {})
      options[:except] ||= :sopra_code
      super
    end
    
    def as_json(options = {})
      options[:except] ||= :sopra_code
      super
    end
    
    def applicant
      housing_request&.main_applicant
    end
    
    def applicant_id
      applicant&.id
    end
    
    def update_offer_applicated_nb
      return unless housing_offer
      applicated_nb = Hermes::HousingApplication.where(
        is_active: true, 
        housing_offer_id: housing_offer_id,
      ).count
      housing_offer.set(updated_at: DateTime.current, applicated_nb: applicated_nb)
      housing_offer.handle_unpublished_status
      housing_offer.save if housing_offer.changed?
    end

    def set_als_id
      self.als_id = housing_request&.als_id
    end

    def als_id
      value = super
      return value if value.present?

      if housing_request
        value = housing_request.als_id
        set(als_id: value, updated_at: DateTime.current)
        self.als_id = value
        value
      end
    end

    def handle_status_changed
      applicant_cancel_application if disabled_or_deleted_status?
     
      if status&.upcase == 'RETENUE EN CAL'
        applicant.completed_requests_nb ? (applicant.completed_requests_nb += 1) : (applicant.completed_requests_nb = 1)
        applicant.save
        housing_request.update_applicant_eligbility
      end
      
      code_to_send = nil
      if status_changed?
        code_to_send = Hermes::Statuses::HousingApplication.sopra_code(status)
        create_public_share if status == 'TRANSMIS BAILLEUR'
      end
      if visit_status_changed?
        case visit_status
        when 'VISITE PROPOSEE'
          code_to_send = "VISPRO"
        when 'VISITE EFFECTUEE'
          code_to_send = "VISEFF"
        end
      end
      
      return unless code_to_send && !code_to_send.blank?
      send_updated_status_task(code_to_send)
    end

    def send_updated_status_task(code_to_send)
      so = Hermes::BoClient::UpdateStatusTask.new({
        housing_application_id: id,
        housing_request_id: housing_request_id,
        housing_offer_id: housing_offer_id,
        code_evenement: code_to_send,
        date_evenement: last_status_update_date || Date.current,
        date_survenance_evenement: date_event,
        rang: rank 
      })
      Rails.env.test? ? true : so.execute
    end 
    
    def handle_visit_conducted_date
      if visit_conducted_date
        if visit_conducted_date.past?
          self.set(visit_status: 'VISITE EFFECTUEE', updated_at: DateTime.current) if visit_status.blank?
        else
          self.set(visit_status: 'VISITE PROPOSEE', updated_at: DateTime.current) if visit_status.blank?
        end
      end
    end

    def update_user_id_from_request
      self.user_id = housing_request.user_id
    end 
    
    def before_save_actions
      check_order_status if status_changed?
      update_application_score
      update_cotations
      update_business_cotation
      update_user_id_from_request if housing_request_id_changed? && housing_request_id
      change_is_active if status_changed? || visit_status_changed?
      update_reader_writers_from_offer if housing_offer_id_changed?
      set_als_id if housing_request_id_changed?
      true
    end

    def check_order_status
      return if status_was.blank? || status.blank?

      status_keys = Hermes::Statuses::HousingApplication.order_status.keys  
      index_status_w = status_keys.include?(status_was) && Hermes::Statuses::HousingApplication.order_status[status_was] 
      index_status = status_keys.include?(status) && Hermes::Statuses::HousingApplication.order_status[status]

      if index_status_w && index_status && index_status <= index_status_w
        self.status = status_was
        self.date_event = date_event_was if date_event_was
        self.last_status_update_date = last_status_update_date_was if last_status_update_date_was
        self.rank = rank_was if rank_was
      end 
    end 

    def update_application_score
      self.profile_score = housing_request.profile_score
    end 

    def update_business_cotation
      self.business_cotation = housing_request.business_cotation
    end 
    
    def update_reader_writers_from_offer
      old_offer = Hermes::HousingOffer.find(housing_offer_id_was) if housing_offer_id_was
      
      if housing_offer_id_was && old_offer
        self.readers = (self.readers || []) - (old_offer.readers || [])
        self.writers = (self.writers || []) - (old_offer.writers || [])
      end
      
      if housing_offer
        self.readers = (self.readers || []) + (housing_offer.readers || [])
        self.writers = (self.writers || []) + (housing_offer.writers || [])
      end
    end
    
    def change_is_active
      self.is_active = !inactive_status?
    end
    
    def before_validation_actions
      if status_changed?
        self.status_updated_at = DateTime.current
        self.sopra_code = Hermes::Statuses::HousingApplication.sopra_code(status)
        #  self.status = Hermes::Statuses::HousingApplication.correspondence(status || '')
      end
      
      if visit_status_changed?
        self.visit_status_updated_at = DateTime.current
        self.visit_status = I18n.transliterate(visit_status || '').upcase
      end
      update_right
    end
    
    def reprise_update_right
      apps = Hermes::HousingApplication.where(is_active: true)
      apps.length
      apps.batch_size(5).each_with_index do |app, i|
        if app.housing_request
          app.update_right
          app.save
        end
      
      end
    end
    
    def update_right
      lessor_user_id = housing_offer&.lessor&.user_id&.to_s
      main_applicant = housing_request&.main_applicant
      reader_inst = [self, housing_request, main_applicant]
      if lessor_user_id && main_applicant
        if disabled_or_deleted_status?
          reader_inst.each do |inst|
            inst.pull(readers: lessor_user_id)
          end
          Hermes::Applicant.where(:id.in => housing_request.applicant_ids).pull(readers: lessor_user_id)
        else
          reader_inst.each do |inst|
            inst.add_to_set(readers: lessor_user_id)
          end
          Hermes::Applicant.where(:id.in => housing_request.applicant_ids).add_to_set(readers: lessor_user_id)
        end
        reader_inst.each do |inst|
          inst.add_to_set(readers: main_applicant.readers)
        end
      end
    end
    
    def only_one_active_application
      return unless housing_offer_id || housing_request_id
      Hermes::HousingApplication.where({
        :is_active => true,
        :housing_offer_id => housing_offer_id,
        :housing_request_id => housing_request_id,
        :id.ne => id
      }).each do |a|
        a.update(is_active: false, status: 'ANNULEE')
      end
    end
    
    def update_request_active_application_id
      if is_active
        housing_request.set(
          updated_at: DateTime.now,
          active_application_id: id.to_s,
          active_application_housing_offer_id: housing_offer_id.to_s
        )
      elsif housing_request.active_application_id.to_s == id.to_s
        housing_request.set(
          updated_at: DateTime.now,
          active_application_id: nil,
          active_application_housing_offer_id: nil
        )
      end 
    end

    def impact_offer_and_request_status
      return unless status_changed?

      if status == 'LOGE'
        housing_request.update(
          request_status: "SATISFAITE",
          attributed_housing_district: housing_offer.district,
          attributed_housing_district_postal_code: housing_offer.postal_code || housing_offer.insee_code,
        )
      else
        housing_request.manage_entity_status
        housing_request.save if housing_request.entity_status_changed?
      end 
      
      case status
      when 'TRANSMIS BAILLEUR'
        housing_offer.update(
          lessor_status: "EN ATTRIBUTION",
          offer_status: "EN COURS D'ATTRIBUTION"
        )
      when 'LOGE'
        housing_offer.update(
          offer_status: "LOGEMENT PLACE",
          lessor_status: "PLACE"
        )
      end
    end
    
    def after_save_actions
      # Broken tests
      check_status_changed if visit_status_changed? || status_changed?
      
      # housing_request&.with_lock(reload: false, wait: true) do
        housing_request.save
      # end
      notify_applicant
      handle_status_changed if visit_status_changed? || status_changed?
      send_flux_50 if (status_changed? || visit_status_changed?) && disabled_or_deleted_status?
      impact_offer_and_request_status 
      only_one_active_application if is_active
      update_request_active_application_id if is_active_changed?
      update_offer_applicated_nb if is_active_changed?
    end
    
    def after_actions
      handle_visit_conducted_date
    end

    def update_cotations
      self.cotations = housing_request.reload.cotations
    end 

    def before_create_actions
      update_cotations
    end 
    
    def after_create_actions
      housing_request.set(last_application_date: created_at, updated_at: DateTime.current)
      HandleRequestStateJob.set(wait: 1.days).perform_later(housing_request_id.to_s, 'request_application')
      
      if Rails.env == 'test'
        {exectuted: true, xml: to_xml}
      else
        Hermes::BoClient::ApplicationTask.new(
          {
            housing_application_id: id,
            housing_request_id: housing_request_id,
            housing_offer_id: housing_offer_id,
          }
        ).execute
      end
    end
    
    def after_destroy_actions
      # housing_request.with_lock(reload: false) do
        housing_request.save
      # end
      update_offer_applicated_nb
    end
    
    def disabled_or_deleted_status?
      Hermes::Statuses::HousingApplication.cancel_status.include?(self.status) || 
      Hermes::Statuses::HousingApplication.inactive_visit_status.include?(self.visit_status)
    end

    def inactive_status?
      Hermes::Statuses::HousingApplication.inactive_status.include?(self.status) || 
      Hermes::Statuses::HousingApplication.inactive_visit_status.include?(self.visit_status)
    end 

    def send_flux_50
      so = Hermes::BoClient::CancelStatusTask.new({
                                                    housing_application_id: id,
                                                    housing_offer_id: housing_offer_id,
                                                    housing_request_id: housing_request_id,
                                                    status_updated_at: status_updated_at || DateTime.now
                                                  })
      if Rails.env.test?
        #{exectuted: true, xml: so.to_xml}
        so.execute
        true
      else
        so.execute
      end
    end 
    
    def applicant_cancel_application
      housing_request.update_applicant_eligbility
      housing_request.save
    end
    
    def system_cancel_application!
      set(status: 'ANNULEE', is_active: false, updated_at: DateTime.current)
      housing_request.update_applicant_eligbility
      housing_request.cancel_request
    end

    def create_public_share
      return nil unless housing_offer.lessor
      Hermes::HousingRequestShare.find_or_create_by(lessor_id: housing_offer.lessor.id, housing_request_id: housing_request.id)
    end
    
    def hr_offers_for_notification
      housing_request.eligibles_offers.where({
        reserved: false,
        offer_status:"LOGEMENT PUBLIE",
        lessor_status: "PUBLIE",
      }).order_by(date_publication_start: :desc).limit(6).pluck(:id).map(&:to_s)
    end

    def notify_applicant
      if status_changed?
        case status.upcase
        when 'RETENUE EN CAL'
          # notify lessor
          # notify(key: NOTIFICATION_KEYS[:housing_application][:lessor_accepted])
        when 'LOGE'
          # notify(key: NOTIFICATION_KEYS[:housing_application][:lessor_accepted], channels: ['email'])
        when 'TRANSMIS BAILLEUR'
          # notify(key: NOTIFICATION_KEYS[:housing_application][:lessor_transferred], channels: ['email'])
        when 'REFUSEE EN CAL', 'NON RETENUE'
          # notify(key: NOTIFICATION_KEYS[:housing_application][:lessor_denied], channels: ['email'])
          # notify(key: NOTIFICATION_KEYS[:housing_application][:cal_denied], channels: ['email'])
          housing_request.user.notify(
            key: NOTIFICATION_KEYS[:user][:application_refusal],
            channels: ['email'],
            datas: {
              offer_ids: hr_offers_for_notification, 
              first_name: housing_request.requester_firstname,
              last_name: housing_request.requester_lastname , 
              external_ref: housing_offer.external_ref,
              city: housing_offer.city,
              applicated_at:(applicated_at || created_at)&.strftime("%d/%m/%Y")
            }
          )
        when 'REPRIS BAILLEUR'
          housing_request.user.notify(
            key: NOTIFICATION_KEYS[:user][:application_offer_taken_back],
            channels: ['email'],
            datas: {
              offer_ids: hr_offers_for_notification, 
              first_name: housing_request.requester_firstname,
              last_name: housing_request.requester_lastname , 
              external_ref: housing_offer.external_ref,
              city:housing_offer.city,
              applicated_at:(applicated_at || created_at)&.strftime("%d/%m/%Y")
            }
          )
        when 'EN ATTENTE'
          # notify(key: NOTIFICATION_KEYS[:housing_application][:pending], channels: [])
        when 'INSCRIPTION CAL'
          # notify(key: NOTIFICATION_KEYS[:housing_application][:subscribed], channels: [])
        when 'DESISTEMENT SALARIE AVANT CAL', 'DESISTEMENT APRES CAL', 'DESISTEMENT'
          # notify(key: NOTIFICATION_KEYS[:housing_application][:applicant_cancelled], channels: [])
        else
          nil
        end
      end
    end
  end
end
