require 'rails_helper'

RSpec.describe Hermes::EntityCode, type: :model do
  let!(:entity) { create :entity }
  let!(:applicant) { create :applicant }
  let!(:entity_code) { create :entity_code, entity_id: entity.id, applicant_id: applicant.id, code: 'ABC' }

  describe 'for structural tests' do
    it 'has a valid factory' do
      expect(FactoryBot.build(:entity_code)).to be_valid
    end
  end

  describe 'validations' do
    describe 'uniqueness code' do
      let(:new_code) { build :entity_code, code: 'ABC' }
      it 'raises error' do
        expect(new_code).to_not be_valid
        expect { new_code.save! }.to raise_error(Mongoid::Errors::Validations)
      end
    end

    describe 'uniqueness scope applicant_id/entity_id' do
      let(:new_code) { build :entity_code, entity_id: entity.id, applicant_id: applicant.id }

      context 'when existed code not expired' do
        it 'raises error' do
          expect(new_code).to_not be_valid
          expect { new_code.save! }.to raise_error(Mongoid::Errors::Validations)
        end
      end

      context 'when existed code expired' do
        before { entity_code.set(expire_at: Time.zone.now.yesterday) }
        it 'succeeds' do
          expect(new_code).to be_valid
          expect { new_code.save! }.to change { Hermes::EntityCode.count }.by 1
        end
      end
    end

    describe 'expiration_date' do
      let(:new_code) { build :entity_code, expire_at: Time.zone.now.yesterday }
      it 'raises error' do
        expect(new_code).to_not be_valid
        expect { new_code.save! }.to raise_error(Mongoid::Errors::Validations, /La date d\'expiration doit être future/)
      end
    end
  end

  describe 'callbacks' do
    describe "#set_code" do
      context 'when no code specified' do
        let(:new_code) { build :entity_code }
        before do
          allow(new_code).to receive(:set_code).and_call_original
          allow(SecureRandom).to receive(:alphanumeric).with(8).and_return 'abcd1234'
        end
        it 'sets upcased random code' do
          expect do
            new_code.save
            expect(new_code).to have_received(:set_code).once
          end.to change { new_code.code }.from(nil).to 'ABCD1234'
        end
      end

      context 'when code is specified' do
        let(:new_code) { build :entity_code, code: 'Specified_CODE' }
        before do
          allow(new_code).to receive(:set_code).and_call_original
        end
        it 'is not called' do
          new_code.save!
          expect(new_code).to_not have_received(:set_code)
          expect(new_code.reload.code).to eq 'Specified_CODE'
        end
      end
    end
  end
end